﻿Public Class ThreadManager
    Implements IDisposable

#Region "Singleton"
    Private Shared _ThreadManager As ThreadManager
    Private Shared ReadOnly mylock As New Object()

    Public Shared Function GetInstance() As ThreadManager
        SyncLock (mylock)
            If _ThreadManager Is Nothing Then
                _ThreadManager = New ThreadManager
            End If
            Return _ThreadManager
        End SyncLock
    End Function
#End Region

#Region "Variable privé"
    Private Task_List As Queue(Of System.Threading.ThreadStart)

    Private mMaxThread As Integer = 4
    Private Thread_List() As Threading.Thread

    Private Thread_Process As Threading.Thread
    Private mWait_StartProcess As Boolean = True

    Private mMutex As Threading.Mutex

    Private Threads_Priority As Threading.ThreadPriority = Threading.ThreadPriority.Normal
#End Region

#Region "Constructeur"
    Private Sub New()
        ReDim Thread_List(MaxThread - 1)
        Thread_Process = New Threading.Thread(AddressOf Process)
        Thread_Process.Name = "thread process" & CInt(Rnd() * 255)
        Task_List = New Queue(Of System.Threading.ThreadStart)
        mMutex = New Threading.Mutex(False, "thread_mutex")
    End Sub
#End Region

    ''' <summary>
    ''' défini le maximum de thread
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property MaxThread() As Integer
        Get
            Return mMaxThread
        End Get
        Set(ByVal value As Integer)
            mMaxThread = value
            ReDim Preserve Thread_List(mMaxThread - 1)
        End Set
    End Property

    ''' <summary>
    ''' Ajoute une fonction à la liste
    ''' </summary>
    ''' <param name="threadfunction">Fonction que vous souhaitez ajouter à la liste: écrivez "AddressOf MaFonction"</param>
    ''' <param name="CanDuplicate">Indique si cette fonction peut être ajouté plusieurs fois à la liste</param>
    ''' <remarks></remarks>
    Public Sub Add(ByVal threadfunction As System.Threading.ThreadStart, ByVal CanDuplicate As Boolean)
        If CanDuplicate = False Then
            If Task_List.Contains(threadfunction) Then Exit Sub
        End If

        mMutex.WaitOne()
        Task_List.Enqueue(threadfunction)
        mMutex.ReleaseMutex()

        If Not mWait_StartProcess Then Start()
    End Sub

    ''' <summary>
    ''' fonction du thread principal qui distribut les taches aux threads disponible
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub Process()
        Do While Task_List.Count > 0
            For mInc0 As Integer = 0 To mMaxThread - 1
                If Task_List.Count > 0 Then
                    If (Thread_List(mInc0) Is Nothing) OrElse (Thread_List(mInc0).IsAlive = False) Then

                        mMutex.WaitOne()
                        Thread_List(mInc0) = New Threading.Thread(Task_List.Dequeue)
                        mMutex.ReleaseMutex()
                        Thread_List(mInc0).Priority = Threads_Priority
                        Thread_List(mInc0).Start()

                    End If
                End If
            Next

        Loop

    End Sub

    ''' <summary>
    ''' Démarre le traitement des taches
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub Start()
        If Thread_Process Is Nothing OrElse Thread_Process.IsAlive = False Then
            Thread_Process = New System.Threading.Thread(AddressOf Me.Process)
            Thread_Process.Start()
        End If
    End Sub

    ''' <summary>
    ''' Le thread principal attends d'être démarré au lieu de commencé dès l'ajout d'une tache
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property Process_StartWait() As Boolean
        Get
            Return mWait_StartProcess
        End Get
        Set(ByVal value As Boolean)
            mWait_StartProcess = value
        End Set
    End Property

    ''' <summary>
    ''' Indique si le thread principal est Actif
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property Process_Alive() As Boolean
        Get
            Return Thread_Process.IsAlive
        End Get
    End Property

    ''' <summary>
    ''' Indique le nombre de thread actif
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property ThreadRun() As Integer
        Get
            Dim total As Integer
            For mInc1 As Integer = 0 To mMaxThread - 1
                If Not Thread_List(mInc1) Is Nothing Then
                    If Thread_List(mInc1).IsAlive Then total += 1
                End If
            Next
            Return total
        End Get
    End Property

    ''' <summary>
    ''' Indique le nombre de tache total à traiter
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property TaskCounts() As Integer
        Get
            If Not Task_List Is Nothing Then
                Return ThreadRun + Task_List.Count
            Else
                Return 0
            End If
        End Get
    End Property

    ''' <summary>
    ''' indique s'il reste encore des taches
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property HaveTask() As Boolean
        Get
            If TaskCounts > 0 Then Return True
        End Get
    End Property

    ''' <summary>
    ''' indique le nombre de tache non traité
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property TaskCount() As Integer
        Get
            If Not Task_List Is Nothing Then
                Return Task_List.Count
            Else
                Return 0
            End If
        End Get
    End Property


#Region "Destructeur"
    Private disposedValue As Boolean = False        ' Pour détecter les appels redondants

    ' IDisposable
    Protected Overridable Sub Dispose(ByVal disposing As Boolean)
        If Not Me.disposedValue Then
            If disposing Then
                ' TODO : libérez des ressources managées en cas d'appel explicite
                Task_List.Clear()
                For i As Integer = 0 To mMaxThread - 1
                    If Not Thread_List(i) Is Nothing Then
                        If Thread_List(i).IsAlive = True Then
                            Thread_List(i).Abort()
                        End If
                        Thread_List(i) = Nothing
                    End If
                Next
                If Thread_Process.IsAlive = True Then
                    Thread_Process.Abort()
                End If
                Thread_Process = Nothing
            End If
            mMutex.Close()
            mMutex = Nothing
        End If
        Me.disposedValue = True
    End Sub

#Region " IDisposable Support "
    ' Ce code a été ajouté par Visual Basic pour permettre l'implémentation correcte du modèle pouvant être supprimé.
    Public Sub Dispose() Implements IDisposable.Dispose
        ' Ne modifiez pas ce code. Ajoutez du code de nettoyage dans Dispose(ByVal disposing As Boolean) ci-dessus.
        Dispose(True)
        GC.SuppressFinalize(Me)
    End Sub
#End Region
#End Region

End Class
